// $Id: OzList.h,v 1.2 2011/09/25 14:00:45 leavens Exp leavens $
// Lists based on the OO paradigm, as described in the paper:
// William R. Cook. "Object-Oriented Programming Versus Abstract Data Types",
// Foundations of Object-Oriented Languages, REX School/Workshop, May/June 1990,
// "Springer-Verlag", 1991, LNCS 489, pages 151-178.
// AUTHOR: Gary T. Leavens

#if !defined(_OZLIST_H)
#define _OZLIST_H
template <typename T>
class OzList { // An abstract class, lists of elements of type T
public:
  // is this list empty?
  virtual bool isEmpty() const = 0;

  // return the first element in the list
  virtual T head() const = 0;

  // return a list containing all but the first element
  virtual OzList<T> const* tail() const = 0;
};

class Empty{}; // a class of exceptions 

template <typename T>
class OzEmptyList : public OzList<T> { // The empty list of type T
public:
  OzEmptyList() {}
  bool isEmpty() const { return true; }
  T head() const {
    throw(Empty());
  }
  OzList<T>* tail() const {
    throw(Empty());
  }
};

template <typename T>
class OzCons : public OzList<T> {  // non-empty lists of type T
protected:
  const T contents;
  const OzList<T> * next_cell;
public:
  OzCons(T x, OzList<T> const* l)
    : contents(x), next_cell(l) {};
  bool isEmpty() const { return false; }
  T head() const { return contents; }
  OzList<T> const* tail() const {
    return next_cell;
  }
};

#include <iostream>
template <typename T>
extern
std::ostream& operator << (std::ostream & out,
			   const OzList<T>& lst);
#endif
